# coding: utf-8

u"""Клиентская часть модуля."""

from uuid import uuid4

from spyne_smev.client import Client
from suds.xsd.doctor import Import, ImportDoctor

from .validation import (
    BaseValidator,
    PutDocumentValidator,
    GetDocumentResultValidator
)
from suds import WebFault


class FsinClient(object):

    u"""Клиент для запросов во ФСИН."""

    TYPE_CODE = 'GSRV'
    EXCHANGE_TYPE = '2'
    TEST_MSG = 'SMEVTEST'

    def __init__(self, wsdl_url, certificate_path, private_key_path,
                 private_key_pass, proxy):
        u"""Проверка параметров и создание клиента для отправки запросов."""
        self.last_sent = None
        self.last_received = None

        assert isinstance(wsdl_url, basestring), type(wsdl_url)
        assert isinstance(certificate_path, basestring), type(certificate_path)
        assert isinstance(private_key_path, basestring), type(private_key_path)
        assert isinstance(private_key_pass, basestring), type(private_key_pass)

        # Принудительный импорт схемы
        schema_url = 'http://www.red-soft.biz/schemas/shgate/common/2011/0.5'
        schema_import = Import(schema_url)
        schema_doctor = ImportDoctor(schema_import)

        # Внутренний клиент для взаимодействия со СМЭВ.
        self.smev_client = Client(
            url=wsdl_url,
            private_key_path=private_key_path,
            private_key_pass=private_key_pass,
            certificate_path=certificate_path,
            proxy=proxy,
            doctor=schema_doctor
        )

    class BaseRequest(object):

        u"""Базовый класс запросов."""

        STATUS = NotImplemented

        def __init__(self):
            self.params = None
            self.smev_client = None
            self.validator = BaseValidator()
            self.request_type = self.__class__.__name__

        def set_params(self, msg, params):
            u"""Запись информации в сообщение.

            :param msg: сообщение suds.
            """
            smev_sender_code = params.pop('SMEV_SENDER_CODE')
            smev_sender_name = params.pop('SMEV_SENDER_NAME')

            # Код/Название системы-инициатора взаимодействия (потребителя)
            msg.Message.Sender.Code = smev_sender_code
            msg.Message.Sender.Name = smev_sender_name

            # Код/Название системы-получателя сообщения (поставщика)
            msg.Message.Recipient.Code = params.pop('FSIN_SMEV_CODE')
            msg.Message.Recipient.Name = params.pop('FSIN_SMEV_NAME')

            # Код/Название системы, инициировавшей цепочку из нескольких
            # запросов-ответов, объединенных единым процессом
            msg.Message.Originator.Code = smev_sender_code
            msg.Message.Originator.Name = smev_sender_name

            msg.Message.TypeCode = FsinClient.TYPE_CODE
            msg.Message.Status = self.STATUS

            # Дата и время формирования запроса
            msg.Message.Date = params.pop('MessageDate')

            msg.Message.ExchangeType = FsinClient.EXCHANGE_TYPE
            msg.Message.ServiceCode = params.pop('SERVICE_CODE')

        def execute(self, smev_client, params):
            u"""Выполнение запроса.

            :param smev_client: клиент spyne-smev.
            :type smev_client: :class:`spyne_smev.client.Client`
            :param dict params: Информация, передаваемая в сообщении.
            """
            assert isinstance(smev_client, Client)

            self.validator.validate(params)

            self.smev_client = smev_client

            msg = smev_client.factory.create(self.request_type)

            self.set_params(msg, params)

            service_request = getattr(
                smev_client.service, self.request_type
            )

            # Перехватываем WebFault, чтобы добавить в него последнее
            # отправленное сообщение
            try:
                result = service_request(msg.Message, msg.MessageData)
            except WebFault as exc:
                exc.last_sent = smev_client.last_sent()
                raise exc

            last_sent = unicode(smev_client.last_sent())
            last_received = unicode(smev_client.last_received())

            return result, last_sent, last_received

    class PutDocument(BaseRequest):

        u"""Запрос bookRequest."""

        STATUS = 'REQUEST'

        def __init__(self):
            u"""Задает валидатор для этого типа запроса."""
            super(FsinClient.PutDocument, self).__init__()
            self.validator = PutDocumentValidator()

        def set_params(self, msg, params):
            super(FsinClient.PutDocument, self).set_params(msg, params)

            msg.Message.CaseNumber = str(uuid4())

            msg.MessageData.AppData.IQueryPfrPensionWar = (
                self.smev_client.factory.create(
                    '{http://www.red-soft.biz/schemas/shgate/common/2011/0.5}'
                    'IQueryPfrPensionWar'
                )
            )

            msg.MessageData.AppData.IQueryPfrPensionWar.senderCode = (
                params.pop('SENDER_CODE')
            )

            for param, val in params.iteritems():
                setattr(
                    msg.MessageData.AppData.IQueryPfrPensionWar,
                    param,
                    val
                )

    class GetDocumentResult(BaseRequest):

        u"""Запрос getResult."""

        STATUS = 'PING'

        def __init__(self):
            u"""Задает валидатор для этого типа запроса."""
            super(FsinClient.GetDocumentResult, self).__init__()
            self.validator = GetDocumentResultValidator()

        def set_params(self, msg, params):
            u"""Заполняет раздел Message.

            :param msg: раздел Body.getResult
            :param dict params: параметры запроса
            """
            super(FsinClient.GetDocumentResult, self).set_params(msg, params)

            msg.Message.CaseNumber = params['CaseNumber']
            msg.Message.RequestIdRef = params['RequestIdRef']
            msg.Message.OriginRequestIdRef = params['OriginRequestIdRef']

    def create_request(self, request_type):
        u"""Возвращает инстанс запроса в зависимости от типа запроса.

        :param str request_type: Тип запроса (PutDocument, GetDocumentResult).
        """
        assert isinstance(request_type, basestring), type(request_type)

        try:
            clss = getattr(self, request_type)
        except AttributeError:
            raise NotImplementedError(
                u'`{0}` is not implemented.'.format(request_type)
            )
        else:
            return clss()

    def request(self, request_type, params):
        u"""Выполнение запроса.

        :param str request_type: Требуемый запрос (bookRequest, getResult).
        :param dict params: Информация, передаваемая в сообщении.
        """
        request = self.create_request(request_type)

        result, last_sent, last_received = request.execute(
            self.smev_client, params
        )

        self.last_sent = last_sent
        self.last_received = last_received

        return result
